/*

    File: ntfs.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
//#define NTFS_DEBUG 1

#define NTFS_GETU8(p)      (*(__u8*)(p))
#define NTFS_GETU16(p)     (*(__u16*)(p))
#define NTFS_GETU32(p)     (*(__u32*)(p))
#define NTFS_GETU64(p)     (*(__u64*)(p))
 
#include <stdio.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "intrface.h"
#include "ntfs.h"
#include "fnctdsk.h"
#include "lang.h"
#include "hdaccess.h"
int ntfs_get_attr(char *attr,t_diskext partition);
int dir_ntfs_aux(WINDOW *window,t_param_disk disk_car, t_diskext partition, const struct ntfs_boot_sector*ntfs_header);
int ntfs_dir_attr(WINDOW *window,char *attr,t_diskext partition);

int check_NTFS(t_param_disk disk_car,t_diskext partition,const int debug,const int dump_ind)
{
  uchar buffer[0x200];
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba))
  { return 1; }
  if(test_NTFS(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind))
    return 1;
  set_NTFS_info(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind);
  return 0;
}

int recover_NTFS(t_param_disk disk_car, struct ntfs_boot_sector*ntfs_header,t_diskext partition,const int debug, const int dump_ind)
{
  if(test_NTFS(disk_car,ntfs_header,partition,debug,dump_ind))
    return 1;
  partition->part_size=ntfs_header->sectors_nbr+1;
  partition->part_type=P_NTFS;
  /* Prb for backup sector : partition->lba isn"t correct yet */
  set_NTFS_info(disk_car,ntfs_header,partition,debug,dump_ind);
  return 0;
}

int set_NTFS_info(t_param_disk disk_car, struct ntfs_boot_sector*ntfs_header,t_diskext partition,const int debug, const int dump_ind)
{
  return set_NTFS_part_name(disk_car,partition,ntfs_header);
}

int test_NTFS(t_param_disk_cst disk_car,struct ntfs_boot_sector*ntfs_header, t_diskext partition,const int debug, const int dump_ind)
{
  char*buffer=(char*)ntfs_header;
  if(ntfs_header->marker==0xAA55)
  {
    if(memcmp(ntfs_header->system_id,"NTFS",4)==0)
    {
      if(debug||dump_ind)
      {
	ecrit_rapport("\nNTFS at %lu/%lu/%lu\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      }
      if(dump_ind)
	dump(stdscr,buffer,0x200);
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}

int set_NTFS_part_name(t_param_disk disk_car, t_diskext partition, const struct ntfs_boot_sector*ntfs_header)
{
  unsigned char *buffer;
  unsigned char *attr;
  loff_t mft_pos=partition->lba+ntfs_header->reserved+ntfs_header->mft_cluster*ntfs_header->cluster_size;
  int mft_recordsize;
  partition->name[0]='\0';
  if(ntfs_header->clusters_record>0)
    mft_recordsize=ntfs_header->cluster_size*ntfs_header->clusters_record;
  else
    mft_recordsize=1<<(-ntfs_header->clusters_record);
  buffer=(unsigned char *)MALLOC(mft_recordsize*0x200);
#ifdef NTFS_DEBUG
  ecrit_rapport("NTFS cluster size = %ld\n",ntfs_header->cluster_size);
  ecrit_rapport("NTFS MFT_record_size = %d\n",mft_recordsize);
  ecrit_rapport("NTFS MFT cluster = %ld\n",ntfs_header->mft_cluster);
#endif
  if(hd_read_LBA(disk_car,mft_recordsize, buffer, mft_pos))
  {
    ecrit_rapport(m_ROOT_CLUSTER_RERR);
    FREE(buffer);
    return 1;
  }
  attr=buffer;
  while(attr<(buffer+mft_recordsize*0x200))
  {
    ntfs_get_attr(attr,partition);
#ifdef NTFS_DEBUG
    dump(stdsrc,attr,NTFS_GETU32(attr + 0x1C));
#endif
    if(!NTFS_GETU32(attr + 0x1C))
    {
      FREE(buffer);
      return 1;
    }
    attr+= NTFS_GETU32(attr + 0x1C);
  }
  FREE(buffer);
  return 0;
}

int ntfs_get_attr(char *attr,t_diskext partition)
{
	/* location of first attribute */
	char *it= attr + NTFS_GETU16(attr + 0x14);
	int type;
	int len;
	/* Only check for magic DWORD here, fixup should have happened before */
	if(memcmp(attr,"FILE",4)) return 0;
	do
	{
	  int offset=NTFS_GETU16(it+0x14);
	  char *data=it+offset;
	  type=NTFS_GETU32(it);
	  len=NTFS_GETU16(it+4);
	  if(len==0)
	    return 0;
	  /* We found the attribute type. Is the name correct, too? */
#ifdef NTFS_DEBUG
	  ecrit_rapport("pos=%p type %x data %p\n",it-attr,type,data-attr);
	  dump(stdscr,data,len);
#endif
	  switch(type)
	  {
#ifdef NTFS_DEBUG
	    case 0x30:
	      {
		char *filename_data=data;
		int namelen=NTFS_GETU8(filename_data+0x40);
		char *name_it;
		for(name_it=filename_data+0x42;namelen; name_it+=2,namelen--)
		  ecrit_rapport("%c",*name_it);
	      }
	      break;
#endif
	    case 0x60:
	      {
		char *volumename_data=data;
		int namelen=len-offset;
		char *name_it;
		char *dest=partition->name;
		for(name_it=volumename_data;(namelen>0) && *name_it; name_it+=2,namelen--)
		  *dest++=*name_it;
//		  ecrit_rapport("%c",*name_it);
	      }
	      return 1;
	  }
	  it+=len;
	}while(type!=-1); /* attribute list end with type -1 */
	return 0;
}

int dir_ntfs(WINDOW *window,t_param_disk disk_car,t_diskext partition)
{
  uchar buffer[0x200];
  int debug=0;
  int dump_ind=0;
  aff_buffer(BUFFER_RESET,"Q");
  wmove(window,5,0);
  aff_part(window,1,disk_car,partition);
  aff_part_rapport(1,disk_car,partition);
  ecrit_rapport("\n");
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba))
  { return 1; }
  if(test_NTFS(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind))
  {
    aff_buffer(BUFFER_ADD,"Invalid NTFS data\n");
  }
  else
    dir_ntfs_aux(window,disk_car,partition,(struct ntfs_boot_sector*)&buffer);
  aff_buffer(BUFFER_DISPLAY,"Q",window);
  return 0;
}

int dir_ntfs_aux(WINDOW *window,t_param_disk disk_car, t_diskext partition, const struct ntfs_boot_sector*ntfs_header)
{
  unsigned char *buffer;
  unsigned char *attr;
  loff_t mft_pos=partition->lba+ntfs_header->reserved+ntfs_header->mft_cluster*ntfs_header->cluster_size;
  int mft_recordsize;
  partition->name[0]='\0';
  if(ntfs_header->clusters_record>0)
    mft_recordsize=ntfs_header->cluster_size*ntfs_header->clusters_record;
  else
    mft_recordsize=1<<(-ntfs_header->clusters_record);
  buffer=(unsigned char *)MALLOC(mft_recordsize*0x200);
#ifdef NTFS_DEBUG
  ecrit_rapport("NTFS cluster size = %ld\n",ntfs_header->cluster_size);
  ecrit_rapport("NTFS MFT_record_size = %d\n",mft_recordsize);
  ecrit_rapport("NTFS MFT cluster = %ld\n",ntfs_header->mft_cluster);
#endif
  if(hd_read_LBA(disk_car,mft_recordsize, buffer, mft_pos))
  {
    ecrit_rapport(m_ROOT_CLUSTER_RERR);
    FREE(buffer);
    return 1;
  }
//  dump(window,buffer,mft_recordsize*0x200);
  attr=buffer;
  while(attr<(buffer+mft_recordsize*0x200))
  {
    ntfs_dir_attr(window,attr,partition);
    if(!NTFS_GETU32(attr + 0x1C))
    {
      FREE(buffer);
      return 1;
    }
    attr+= NTFS_GETU32(attr + 0x1C);
  }
  FREE(buffer);
  return 0;
}

int ntfs_dir_attr(WINDOW *window,char *attr,t_diskext partition)
{
	/* location of first attribute */
	char *it= attr + NTFS_GETU16(attr + 0x14);
	int type;
	int len;
	/* Only check for magic DWORD here, fixup should have happened before */
	if(memcmp(attr,"FILE",4)) return 0;
	do
	{
	  int offset=NTFS_GETU16(it+0x14);
	  char *data=it+offset;
	  type=NTFS_GETU32(it);
	  len=NTFS_GETU16(it+4);
	  if(len==0)
	    return 0;
	  /* We found the attribute type. Is the name correct, too? */
#ifdef NTFS_DEBUG
	  ecrit_rapport("pos=%p type %x data %p\n",it-attr,type,data-attr);
	  dump(window,data,len);
#endif
	  switch(type)
	  {
	    case 0x30:
	      {
		char *filename_data=data;
		int namelen=NTFS_GETU8(filename_data+0x40);
		char *name_it;
		for(name_it=filename_data+0x42;namelen; name_it+=2,namelen--)
		  aff_buffer(BUFFER_ADD,"%c",*name_it);
		aff_buffer(BUFFER_ADD,"\n");
	      }
	      break;
	  }
	  it+=len;
	}while(type!=-1); /* attribute list end with type -1 */
	return 0;
}

